;;;;  acset-method-test.lisp - Tests for functions in active-set-method.lisp.
;;;;  Copyright (C) 2010  Norbert Paul
;;;;
;;;;  This program is free software: you can redistribute it and/or modify
;;;;  it under the terms of the GNU General Public License as published by
;;;;  the Free Software Foundation, either version 3 of the License, or
;;;;  (at your option) any later version.
;;;;
;;;;  This program is distributed in the hope that it will be useful,
;;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;;  GNU General Public License for more details.
;;;;
;;;;  You should have received a copy of the GNU General Public License
;;;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;;  Contact: norbert.paul@kit.edu

(eval-when (:compile-toplevel :load-toplevel)
  (load "utils")
  (load "active-set-method")
  (require :asdf)
  (asdf:operate 'asdf:load-source-op 'rt))


(defpackage :active-set-tests 
  (:use :active-set-method :cl :utils)
  (:export #:initially-active-test
	   #:kkt-distance-matrix-test
	   #:simplex-intersection-side-test

;; Fixtures
	   #:dist-test-simplex1
	   #:dist-test-simplex2
	   #:dist-test-matrix
	   #:dist-test-location
	   #:acset-fixtures))

(in-package :active-set-tests)


;;; Test if the initially active matrix is correctly computed.
(defun initially-active-matrix-test-fn ()
    (block the-test
         (let* 
	     (
	      (n (random 20))
	      (result (initially-active n))
	      (expected 0d0))
	   (dotimes (i (+ n 2))
	     (dotimes (j (+ n 2))
	       (setf expected (if (and (< i n) (= i j)) 
				  1d0 ; on the diagonal (except for the last two) 
				  0d0 ; anywhere else.
				  ))
	       (when (/= (aref result i j) expected)
		 (print `(:expected ,expected :for (aref result ,i ,j)  
				    :but :found ,(aref result i j))
			*error-output* )
		 (return-from the-test :failure)))))
	 :success))

(rt:deftest 
    initially-active-test 
    (initially-active-matrix-test-fn)
  :success)

;;; The fixture:
(defparameter dist-test-simplex1 ())
(defparameter dist-test-simplex2 ())
(defparameter dist-test-matrix   ())
(defparameter dist-test-location ())
;defparameter



;;; tests if the kkt-distance-matrix is correctly computetd.
;;; This is a very simple test that just computes a random kkt-distance matrix
;;; and then visits each entry, tests which "compartment" it is in and 
;;; compares the found value with the expected value.
;;; Note that it was this test which found the SSE2-bug in cmucl.
(defun kkt-distance-matrix-test-fn ()
    (block the-test
      (let* 
	  (
	   (n (random 10))
	   (m (random 10))
	   (n+m (+ n m))
	   (n+m+1 (1+ n+m))
	   (n+m+2 (1+ n+m+1))
	   (dim (random 10))
	   (coordinates-factory 
	    #'(lambda ()
		(if (zerop (random 2))
		    (coerce (random 10) 'double-float)
		    (coerce (- (random 10)) 'double-float))))
	   (vertex-factory
	    #'(lambda ()
		(let ((pt (make-array dim :element-type 'double-float)))
		  (map-into pt coordinates-factory))))
	   (null-array (make-array 0 :element-type 'double-float))
	   (simplex1 (map-into (if (zerop (random 2))
				   (make-list n)
				   (make-array n))
			       vertex-factory))
	   (simplex2 (map-into (if (zerop (random 2))
				   (make-list m)
				   (make-array m))
			       vertex-factory))
	   (result   (kkt-distance-matrix simplex1 simplex2)))

	(setf dist-test-simplex1 simplex1
	      dist-test-simplex2 simplex2
	      dist-test-matrix   result)


	(dotimes (i n+m+2)
	  (loop :for j :from i :below n+m+2
	     :do

	     (setf dist-test-location (list i j))

	     (let* ((cell (aref result i j))
		    (vti  (cond ((< i  n)     (elt simplex1 i))
				((< i  n+m)   (elt simplex2 (- i n)))
				(T null-array)))
		    (vtj  (cond ((< j  n)     (elt simplex1 j))
				((< j  n+m)   (elt simplex2 (- j n)))
				(T null-array)))
		    (dot (coerce (reduce  #'+ (map 'list #'* vti vtj)) 'double-float))
		    (blasdot (laroutines:ddot (coerce vti '(vector double-float))
					      (coerce vtj '(vector double-float)))) )

	       ;;The compartments of an KKT-distance matrix:
	       ;;    0    n   n+m                 : the starting row/col - indexes
	       ;; [  AA  -AB  c1  c0 ]   AA,  BB  : "diagonal" matrices
	       ;; [ -BA   BB  c0  c1 ]  -AB, -BA  : "lateral" matrices.
	       ;; [  r1   r0   0   0 ]   r1,  r0  : a constantly 1 or 0 row
	       ;; [  r0   r1   0   0 ]   c1,  c0  ; a constantly 1 or 0 column
	       ;;                              0  ; a zero value.

	       (cond ((/= dot blasdot) ; e.g. by an SSE2-bug ;)
		      (return-from the-test `(:dot-inconsistent ,dot ,blasdot)))

		     ((/= cell (aref result j i)) ; therefore j may start from i.
		      (return-from the-test :not-symmetric))

		    ;; Note. #'< #'> and the like return /generalized/ boolean.
		     ((and (< i n+m) (< j n+m)
			    (equivalent (< i n) (< j n))) ; the "diagonal" blocks AA or BB.
		      (when (/= dot cell)
			(return-from the-test `(:not-dot-product ,cell ,dot at (,i ,j)))))

		     ((and (< i n+m) (< j n+m)
			   (equivalent (< i n) (>= j n))) ; the "lateral" blocks
		      (unless (= (- dot) cell)
		        (return-from the-test `(:not-minus-dot-product ,cell ,dot at (,i ,j)))))
		    
		     ((and (< i n) (= j n+m))    ; first constantly 1 column
		      (unless (= 1d0 cell)
		        (return-from the-test :not-affine-1)))

		     ( (and (< i n) (= j n+m+1)) ; second constantly 0 column
		      (unless (= 0d0 cell)
		        (return-from the-test :not-affine-1)))
		 
		     ( (and (<= n i) (< i n+m) (= j n+m+1)) ; second constantly 1 column
		      (unless (= 1d0 cell)
		        (return-from the-test :not-affine-2)))

		     ( (and (<= n i) (< i n+m) (= j n+m)) ; first constantly 0 column
		      (unless (= 0d0 cell)
		        (return-from the-test :not-affine-2)))
		      
		     ( T ; the four zeros
		      (unless (= 0d0 cell)
			(return-from the-test :not-zero-corner))))
	       )))))
	:success )
  
(rt:deftest kkt-distance-matrix-test (kkt-distance-matrix-test-fn) :success)





;;; these are some simple fixtures for simplex-intersection-vertices:

(defparameter acset-fixtures ())
(defun simplex-intersection-vertices-test-fn ()

      (block the-test
	
	(setf acset-fixtures ())
	
	(push '(#(#(0d0) #(2d0))   ; simplex1 0--------2
		#(#(1d0) #(3d0))   ; simplex2     1----------3
		((0 1 2) (1 2 3))  ; result
	        :fixture-1)        ; informative on error
	      acset-fixtures )

	(push '(#(#(0d0 2d0) #(2d0 2d0) #(1d0 0d0))    ; simplex1  0:0,2--------1:2,2
	                                               ;              \        /
	                                               ;               \3:1,1-x-----4:3,1 simplex2
					               ;                \    /
	                                               ;                 2:1,0
		#(#(1d0 1d0) #(3d0 1d0))               ; simplex2
		((0 1 2 3) (1 2 3 4))                  ; result
		:fixture-2 )
	      acset-fixtures )

	(push '(#(#(0d0 2d0) #(2d0 2d0) #(1d0  0d0))    ; simplex1  0:0,2--------1:2,2
	                                                ;              \        /
	                                                ;               \      / 3:2,1---4:3,1 simplex2
                                                        ;                \    /
	                                                ;                 2:1,0
		#(#(2d0 1d0) #(3d0 1d0))                ; simplex2
		() 
		:fixture-3 )
	      acset-fixtures )

	(push '(#(#(0d0 2d0) #(2d0 2d0) #(1d0 0d0))    ; simplex1  0:0,2--------1:2,2
	                                               ;              \        /
	                                               ;    3:-2,1-----x------x---------4:3,1 simplex2
                                                       ;                \    /
	                                               ;                 2:1,0
		#(#(-2d0 1d0) #(3d0 1d0))              ; simplex2
		((0 2 3 4) (1 2 3 4))
		:fixture-4 )
	      acset-fixtures )
	
	(setf acset-fixtures (nreverse acset-fixtures))

	(dolist (fixture acset-fixtures)
	  (let* ((simplex1 (car fixture))
		 (simplex2 (cadr fixture))
		 (expected (caddr fixture))
		 (id       (print (cadddr fixture)))
		 (result   (print (simplex-intersection-vertices simplex1 simplex2))))
	    (terpri)
	    (unless (set-equals result expected :test #'set-equals)
	      (return-from the-test `(,id :failed)))))

      :success )
)

(rt:deftest 
    simplex-intersection-vertices-test 
    (simplex-intersection-vertices-test-fn)
  :success)




;;; these are some simple fixtures for simplex-intersection-edges
;;; Note that this function is obsolete, so the test isn't any more necessary
(defparameter acset-fixtures-edges ())

(defun simplex-intersection-edges-fn ()
  (block the-test

    ;; Simple Test-Cases:
    ;; - Two Triangles in IR^2
    ;;
    ;;                   * 5:(0,3)
    ;;                  / \
    ;;                 /   \
    ;;                / 2345\
    ;;               /   * 2:(0,2)
    ;;              /   / \   \
    ;;             /   /   \   \
    ;;            /   / 234 \   \
    ;;  3:(-1,1) *---x-------x---* 4:(1,1)
    ;;              /0234     \1234
    ;;             /           \
    ;;            /             \
    ;;  0:(-1,0) *---------------* 1:(1,0)
    ;;
    (let ((vertices 
	   (simplex-intersection-vertices
	    #(#(-1d0 0d0) #(1d0 0d0) #(0d0 2d0))
	    #(#(-1d0 1d0) #(1d0 1d0) #(0d0 3d0)))))
      (print (simplex-intersection-edges vertices)))

    ;; - Two Triangles in IR^3:
    ;;
    ;;                      * 5:(0,3,1)
    ;;                     / \
    ;;                    /   \
    ;;                   /     \
    ;;                  /   * 2:(0,2,0)
    ;;                 /   / \   \
    ;;                /   /   x   \
    ;;               /   /   / \   \
    ;;  3:(-1,1,0)  *-------x---\---* 4:(1,1,0)
    ;;                 /         \    
    ;;                /           \
    ;;               /             \
    ;;  0:(-1,0,-1) *---------------* 1:(1,0,1)
    ;;
    (let ((vertices 
	   (simplex-intersection-vertices
	    #(#(-1d0  0d0 -1d0) #(1d0 0d0 1d0) #(0d0 2d0 0d0))
	    #(#(-1d0  1d0  0d0) #(1d0 1d0 0d0) #(0d0 3d0 1d0))
	    1d-13		       ; Note: fails with default 0d0.
	    )))
      (print (simplex-intersection-edges vertices)))


    (let ((vertices 
	   (simplex-intersection-vertices
	    #(#(-1d0  0d0 -1d0) #(1d0 0d0 1d0) #(0d0 2d0 0d0))
	    #(#(-1d0  1d0  0d0) #(1d0 1d0 0d0) #(0d0 3d0 1d0))
	    )))
      (print (simplex-intersection-edges vertices)))

    ;; This shows why the above fails:
    ;; The two Lagrangians are not zero
    (print (laroutines:solve 
	    (kkt-distance-matrix 
	     #(#(1d0 0d0 1d0) #(0d0 2d0 0d0))
	     #(#(-1d0  1d0  0d0) #(1d0 1d0 0d0) #(0d0 3d0 1d0)))

	    (print (kkt-distance-right-hand-side 5))

	    ))

    )
  :no-conditions ; no results are expected yet, It should simply not raise a condition.
)

(rt:deftest 
    simplex-intersection-edges-test 
    (simplex-intersection-edges-test-fn)
  :no-conditions)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; these are some simple fixtures for simplex-intersection-faces
;;; Note that that function, too, is obsolete, so the test isn't any more necessary

(defun simplex-intersection-faces-test-fn ()
  (block the-test

    ;; - Triangle and Tetrahedron in IR^3:
    ;;
    ;;                      * 5:(0,3,1)
    ;;                     /:\
    ;;                    / : \
    ;;                   /  :  \
    ;;                  /   * 2:(0,2,0) 6:(0,2,3)
    ;;                 /  ./ \.  \
    ;;                / . /   x . \
    ;;               /.  /   /.x  .\
    ;;  3:(-1,1,0)  *-------x---\---* 4:(1,1,0)
    ;;                 /         \    
    ;;                /           \
    ;;               /             \
    ;;  0:(-1,0,-1) *---------------* 1:(1,0,1)
    ;;
    (let ((vertices (simplex-intersection-vertices
		     #(#(-1d0  0d0 -1d0) #(1d0 0d0 1d0) #(0d0 2d0 0d0))
		     #(#(-1d0  1d0  0d0) #(1d0 1d0 0d0) #(0d0 3d0 1d0) #(0d0 2d0 3d0))
		     1d-13   ; Note: fails with former default of 0d0.
		     )))
      (print (simplex-intersection-edges vertices))
      (print (simplex-intersection-faces vertices)))

    ;; - Ya Triangle and Tetrahedron in IR^3:
    ;;                 
    ;;                           * 2:(0,2,0)
    ;;                          / \
    ;;                         /   \
    ;;                        /     \
    ;;                       /       \
    ;;                      /         \
    ;;                     /           \
    ;;                    /  5:(0,1,-1) \
    ;;                   /       *       \
    ;;                  /       / \       \
    ;;                 /       /   \       \
    ;;                / 01256 x-----x 01245 \
    ;;               /       /.     .\       \
    ;;              /       / .     . \       \
    ;;          6:(-1,0,1) *-----------* 4:(1,0,1)
    ;;            /         \ .     . /         \
    ;;           /           \.     ./           \
    ;;          /       01236 x-----x 01234       \
    ;;         /               \   /               \
    ;;        /                 \ /                 \
    ;;       /                   *                   \
    ;;      /               3:(0,-1,-1)               \
    ;;     /                                           \
    ;;    /                                             \
    ;;   /                                               \
    ;;  /                                                 \
    ;; *---------------------------------------------------*
    ;; 0:(-2,-2,0)                                         1:(2,-2,0)


    (let* ((vertices (simplex-intersection-vertices
		      #(#(-2d0 -2d0  0d0) #( 2d0 -2d0  0d0) #( 0d0  2d0  0d0))
		      #(#( 0d0 -1d0 -1d0) #( 1d0  0d0  1d0) #( 0d0  1d0 -1d0) #(-1d0  0d0  1d0))
		      1d-13 ; Note: might fail with former defaults of 0d0 and (depsilon).
		      ))
	   (edges    (simplex-intersection-edges vertices))
	   (uedges   (mapcar #'(lambda (ab) (remove-duplicates (safe-sort (flatten ab) #'<)))
			     edges))
	   (faces    (simplex-intersection-faces vertices uedges)))
      (print `(vertices ,vertices))
      (print `(edges    ,edges))
      (print `(uedges   ,uedges))
      (print `(faces    ,faces))
      (print `(ufaces    ,(remove-duplicates 
			   (mapcar #'(lambda (ab) (remove-duplicates (safe-sort (flatten ab) #'<)))
				   faces)
			   :test #'set-equals)))
      (values)
      )
  :no-conditions ; no results are expected yet, It should simply not raise a condition.
))

(rt:deftest 
    simplex-intersection-faces-test 
    (simplex-intersection-faces-test-fn)
  :no-conditions)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; A "test case" -- better, an experimental form -- for simplex-intersection sides:

(defun simplex-intersection-sides-test-fn ()

  (let* ((vertices (simplex-intersection-vertices
		    #(#(-2d0 -2d0  0d0) #( 2d0 -2d0  0d0) #( 0d0  2d0  0d0))
		    #(#( 0d0 -1d0 -1d0) #( 1d0  0d0  1d0) #( 0d0  1d0 -1d0) #(-1d0  0d0  1d0))
		    1d-13 ; Note: might fail with former defaults of 0d0 and (depsilon).
		    )))
    (print `(vertices ,vertices))
    (print `(sides ,(simplex-intersection-sides vertices))))





  :no-conditions
)

(rt:deftest 
    simplex-intersection-sides-test 
    (simplex-intersection-sides-test-fn)
  :no-conditions)




(define-symbol-macro test-1
    ;; this test case fails:
    (active-set-method::CO-TRIANGULATE-CELL '|a| 'A
					    (active-set-method::pack-boundary 
					     '((((A (|b| |b|)) . -1) ((A (|b| |d|)) . -1)
						((A (|a| |a|)) . 1) ((A (|a| |c|)) . 1)
						((A (|c| |c|)) . 1) ((A (|c| |d|)) . 1)
						((A (|a| |a|)) . -1) ((A (|a| |b|)) . -1))
					       ((((|b| |b|) (|b|)) . 0) (((|b| |d|) (|d|)) . 1)
						(((|b| |d|) (|b|)) . -1) (((|a| |c|) (|a|)) . -1)
						(((|a| |c|) (|c|)) . 1) (((|c| |c|) (|c|)) . 0)
						(((|c| |d|) (|d|)) . 1) (((|c| |d|) (|c|)) . -1)
						(((|a| |a|) (|a|)) . 0) (((|a| |b|) (|b|)) . 1)
						(((|a| |b|) (|a|)) . -1)))
					     :test #'equal)
					    '((A)
					      ((|b| |b|) (|b| |d|) (|a| |c|) (|c| |c|) (|c| |d|)
					       (|a| |a|) (|a| |b|))
					      ((|c|) (|a|) (|d|) (|b|)))
					    :TEST #'equal))

(define-symbol-macro test-1-x
    ;; this test case fails:
    (active-set-method::CO-TRIANGULATE-CELL '|x| 'A
					    (active-set-method::pack-boundary 
					     '((((A (|b| |b|)) . -1) ((A (|b| |d|)) . -1)
						((A (|a| |a|)) . 1) ((A (|a| |c|)) . 1)
						((A (|c| |c|)) . 1) ((A (|c| |d|)) . 1)
						((A (|a| |a|)) . -1) ((A (|a| |b|)) . -1))
					       ((((|b| |b|) (|b|)) . 0) (((|b| |d|) (|d|)) . 1)
						(((|b| |d|) (|b|)) . -1) (((|a| |c|) (|a|)) . -1)
						(((|a| |c|) (|c|)) . 1) (((|c| |c|) (|c|)) . 0)
						(((|c| |d|) (|d|)) . 1) (((|c| |d|) (|c|)) . -1)
						(((|a| |a|) (|a|)) . 0) (((|a| |b|) (|b|)) . 1)
						(((|a| |b|) (|a|)) . -1)))
					     :test #'equal)
					    '((A)
					      ((|b| |b|) (|b| |d|) (|a| |c|) (|c| |c|) (|c| |d|)
					       (|a| |a|) (|a| |b|))
					      ((|c|) (|a|) (|d|) (|b|)))
					    :TEST #'equal))



(define-symbol-macro test-2
    ;; this test case works well:
    (active-set-method::CO-TRIANGULATE-CELL '|a| 'A
					    (active-set-method::pack-boundary 
					     '((((A |bb|) . -1) ((A |bd|) . -1)
						((A |aa|) . 1) ((A |ac|) . 1)
						((A |cc|) . 1) ((A |cd|) . 1)
						((A |aa|) . -1) ((A |ab|) . -1))
					       (((|bb| |b|) . 0) ((|bd| |d|) . 1)
						((|bd| |b|) . -1) ((|ac| |a|) . -1)
						((|ac| |c|) . 1) ((|cc| |c|) . 0)
						((|cd| |d|) . 1) ((|cd| |c|) . -1)
						((|aa| |a|) . 0) ((|ab| |b|) . 1)
						((|ab| |a|) . -1)))
					     :test #'equal)
					    '((A)
					      (|bb| |bd| |ac| |cc| |cd| |aa| |ab|)
					      (|c| |a| |d| |b|))
					    :TEST #'equal))

(active-set-method::co-boundaryp
 (mapcar (utils:rcurry #'active-set-method::chain+ :test #'equal)
	 '(() ())
	 '(((((|a| |a| |b|) (|a| |b|)) . -1) (((|a| |c| |d|) (|c| |d|)) . 1)
	    (((|a| |c| |c|) (|c| |c|)) .  1) (((|a| |a| |c|) (|a| |c|)) . 1) 
	    (((|a| |a| |a|) (|a| |a|)) . 0)
	    (((|a| |b| |d|) (|b| |d|)) . -1) (((|a| |b| |b|) (|b| |b|)) . -1)
	    (((|a| |a| |b|) (|a| |b|)) .  1) (((|a| |a| |b|) (|a| |a|)) . -1)
	    (((|a| |c| |d|) (|a| |d|)) . -1) (((|a| |c| |d|) (|a| |c|)) . 1)
	    (((|a| |c| |c|) (|a| |c|)) .  0) (((|a| |a| |c|) (|a| |a|)) . 1)
	    (((|a| |a| |c|) (|a| |c|)) . -1) (((|a| |a| |a|) (|a| |a|)) . 0)
	    (((|a| |b| |d|) (|a| |d|)) .  1) (((|a| |b| |d|) (|a| |b|)) . -1)
	    (((|a| |b| |b|) (|a| |b|)) .  0))
	   ((((|a| |b|) (|b|)) . 1) (((|a| |b|) (|a|)) . -1) (((|c| |d|) (|d|)) . 1) (((|c| |d|) (|c|)) . -1)
	    (((|c| |c|) (|c|)) . 0) (((|a| |c|) (|a|)) . -1) (((|a| |c|) (|c|)) . 1) (((|a| |a|) (|a|)) . 0)
	    (((|b| |d|) (|d|)) . 1) (((|b| |d|) (|b|)) . -1) (((|b| |b|) (|b|)) . 0) (((|a| |c|) (|c|)) . 1)
	    (((|a| |d|) (|d|)) . 1) (((|a| |a|) (|a|)) . 1) (((|a| |b|) (|b|)) . 1)
	    (((|a| |c|) (|a|)) . -1) (((|a| |d|) (|a|)) . -1)
	    (((|a| |a|) (|a|)) . -1) (((|a| |b|) (|a|)) . -1)))))

(active-set-method::co-boundaryp 
 (mapcar (utils:rcurry #'active-set-method::chain+ :test #'equal)
	 '(() ())
	 '((((( |a| |a| |b|) (|a| |b|)) . -1)
	    ((( |a| |c| |d|) (|c| |d|)) . 1)
	    ((( |a| |c| |c|) (|c| |c|)) . 1)
	    ((( |a| |a| |c|) (|a| |c|)) . 1)
	    ((( |a| |a| |a|) (|a| |a|)) . 0)
	    ((( |a| |b| |d|) (|b| |d|)) . -1)
	    ((( |a| |b| |b|) (|b| |b|)) . -1)
	    ((( |a| |a| |b|) ( |a| |b|)) . 1)
	    ((( |a| |a| |b|) ( |a| |a|)) . -1)
	    ((( |a| |c| |d|) ( |a| |d|)) . -1)
	    ((( |a| |c| |d|) ( |a| |c|)) . 1)
	    ((( |a| |c| |c|) ( |a| |c|)) . 0)
	    ((( |a| |a| |c|) ( |a| |a|)) . 1)
	    ((( |a| |a| |c|) ( |a| |c|)) . -1)
	    ((( |a| |a| |a|) ( |a| |a|)) . 0)
	    ((( |a| |b| |d|) ( |a| |d|)) . 1)
	    ((( |a| |b| |d|) ( |a| |b|)) . -1)
	    ((( |a| |b| |b|) ( |a| |b|)) . 0))
	   ((((|a| |b|) (|b|)) . 1) (((|a| |b|) (|a|)) . -1) (((|c| |d|) (|d|)) . 1)
	    (((|c| |d|) (|c|)) . -1) (((|c| |c|) (|c|)) . 0) (((|a| |c|) (|a|)) . -1)
	    (((|a| |c|) (|c|)) . 1) (((|a| |a|) (|a|)) . 0) (((|b| |d|) (|d|)) . 1)
	    (((|b| |d|) (|b|)) . -1) (((|b| |b|) (|b|)) . 0)
	    ((( |a| |c|) (|c|)) . 1) ((( |a| |d|) (|d|)) . 1)
	    ((( |a| |a|) (|a|)) . 1) ((( |a| |b|) (|b|)) . 1)
	    ((( |a| |c|) ( |a|)) . -1)
	    ((( |a| |d|) ( |a|)) . -1)
	    ((( |a| |a|) ( |a|)) . -1)
	    ((( |a| |b|) ( |a|)) . -1))))
 :test #'equal)

(mapc #'active-set-method::write-boundary
	(mapcar (utils:rcurry #'active-set-method::chain+ :test #'equal)
		'(() ())
		'((((( |a| |a| |b|) (|a| |b|)) . -1)
		   ((( |a| |c| |d|) (|c| |d|)) . 1)
		   ((( |a| |c| |c|) (|c| |c|)) . 1)
		   ((( |a| |a| |c|) (|a| |c|)) . 1)
		   ((( |a| |a| |a|) (|a| |a|)) . 0)
		   ((( |a| |b| |d|) (|b| |d|)) . -1)
		   ((( |a| |b| |b|) (|b| |b|)) . -1)
		   ((( |a| |a| |b|) ( |a| |b|)) . 1)
		   ((( |a| |a| |b|) ( |a| |a|)) . -1)
		   ((( |a| |c| |d|) ( |a| |d|)) . -1)
		   ((( |a| |c| |d|) ( |a| |c|)) . 1)
		   ((( |a| |c| |c|) ( |a| |c|)) . 0)
		   ((( |a| |a| |c|) ( |a| |a|)) . 1)
		   ((( |a| |a| |c|) ( |a| |c|)) . -1)
		   ((( |a| |a| |a|) ( |a| |a|)) . 0)
		   ((( |a| |b| |d|) ( |a| |d|)) . 1)
		   ((( |a| |b| |d|) ( |a| |b|)) . -1)
		   ((( |a| |b| |b|) ( |a| |b|)) . 0))
		  ((((|a| |b|) (|b|)) . 1) (((|a| |b|) (|a|)) . -1) (((|c| |d|) (|d|)) . 1)
		   (((|c| |d|) (|c|)) . -1) (((|c| |c|) (|c|)) . 0) (((|a| |c|) (|a|)) . -1)
		   (((|a| |c|) (|c|)) . 1) (((|a| |a|) (|a|)) . 0) (((|b| |d|) (|d|)) . 1)
		   (((|b| |d|) (|b|)) . -1) (((|b| |b|) (|b|)) . 0)
		   ((( |a| |c|) (|c|)) . 1) ((( |a| |d|) (|d|)) . 1)
		   ((( |a| |a|) (|a|)) . 1) ((( |a| |b|) (|b|)) . 1)
		   ((( |a| |c|) ( |a|)) . -1)
		   ((( |a| |d|) ( |a|)) . -1)
		   ((( |a| |a|) ( |a|)) . -1)
		   ((( |a| |b|) ( |a|)) . -1)))))



(defparameter *get-star-test-cellss*     ())
(defparameter *get-star-test-boundaries* ())


(defun get-star-test-fn () (block nil
  (let* ( (cellss      (copy-tree active-set-method::*sample-cube-cells*))
	  (boundaries  (copy-tree active-set-method::*sample-cube-boundary*))
	  (pad         (make-list (length cellss) :initial-element ()))
	  (core-cellss `((,(caar cellss)) . ,(cdr pad)))
	  ;; core-cells = '((a) () () ())
	  result-cellss result-boundaries)
      
      (setf (values result-boundaries result-cellss)
	    (active-set-method::get-star* core-cellss boundaries))

      (active-set-method::printc `(result-bd = ,result-boundaries))
      (active-set-method::printc `(result-cells = ,result-cellss))

    
      (unless (every #'subsetp (append core-cellss pad) (append result-cellss pad))
	(print `(core = ,core-cellss))
	(print `(result = ,result-cellss))
	(return :failed-core-too-big))

      (unless (every #'subsetp (append result-cellss pad) (append cellss pad))
	(return :failed-result-too-big))

      
      

      :success
 
)))


