;;;;  chains-matrices.lisp
;;;;  Chain and matrix utility functions for the package :active-set-method.
;;;;  All destructive functions raise an error. They are deprecated until the 
;;;;  Software is stable
;;;;
;;;;  Copyright (C) 2010  Norbert Paul
;;;;
;;;;  This program is free software: you can redistribute it and/or modify
;;;;  it under the terms of the GNU General Public License as published by
;;;;  the Free Software Foundation, either version 3 of the License, or
;;;;  (at your option) any later version.
;;;;
;;;;  This program is distributed in the hope that it will be useful,
;;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;;  GNU General Public License for more details.
;;;;
;;;;  You should have received a copy of the GNU General Public License
;;;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;;  Contact: norbert.paul@kit.edu


(in-package :active-set-method)

;;; Chains

(defun old-chain+ (a-chain b-chain &key (test #'eql))
  "A binary addition of two chains. 
A chain is an a-list of arbitrary keys each 
associated to a number."

  (let ((a-result ())
	(b-removals ()))
    (dolist (xA a-chain)
      (let* ((a  (car xA))
	     (xB (find a b-chain :test test :key #'car)))
	(if xB
	    (progn (push xB b-removals)
		   (push `(,a . ,(+ (cdr xA) (cdr xB))) a-result))
	    (push xA a-result))))
    (nconc  (nreverse a-result) 
	    (set-difference b-chain b-removals :test #'eq))))

(defun chain+ (a-chain b-chain &key (test #'eql))
  "A binary addition of two chains. 
A chain is an a-list of arbitrary keys each 
associated to a number."

  (let ((result ()))
    (dolist (chain (list a-chain b-chain))
      (dolist (entry chain)
	(let* ((key  (car entry))
	       (result-entry (assoc key result :test test)))
	  (if result-entry
	      (incf (cdr result-entry) (cdr entry)) 
	      (push `(,key . ,(cdr entry)) result)))))
    (nreverse result)))

(defun chain- (a-chain b-chain &key (test #'eql))
  "A binary addition of two chains. 
A chain is an a-list of arbitrary keys each 
associated to a number."

  (let ((result ()))
    (dolist (entry a-chain)
      (let* ((key  (car entry))
	     (result-entry (assoc key result :test test)))
	(if result-entry
	    (incf (cdr result-entry) (cdr entry)) 
	    (push `(,key . ,(cdr entry)) result))))
    (dolist (entry b-chain)
      (let* ((key  (car entry))
	     (result-entry (assoc key result :test test)))
	(if result-entry
	    (decf (cdr result-entry) (cdr entry)) 
	    (push `(,key . ,(- (cdr entry))) result))))
    (nreverse result)))


(defun old-chain- (a-chain b-chain &key (test #'eql))
  "A binary subtraction of one chain from another chain. 
A chain is an a-list of arbitrary keys each 
associated to a number."

  (let ((a-result ())
	(b-removals ()))
    (dolist (xA a-chain)
      (let* ((a  (car xA))
	     (xB (find a b-chain :test test :key #'car)))
	(if xB
	    (progn (push xB b-removals)
		   (push `(,a . ,(- (cdr xA) (cdr xB))) a-result))
	    (push xA a-result))))
    (nconc  (nreverse a-result) 
	    (mapcar #'(lambda (xB)
			(if (zerop (cdr xB))
			    xB
			    `(,(car xB) . ,(- (cdr xB)))))
		    (set-difference b-chain b-removals :test #'eq)))))

(defun chain-dot-product (chain-1 chain-2 &key (test #'eql))
  "Computes the dot product of the two chains. 
Only the keys of chain-1 will be used in the result."
  (mapcan #'(lambda (entry &aux (other (assoc (car entry) chain-2 :test test)))
	    (when other
	      `((,(car entry) . ,(* (cdr entry) (cdr other))))))
	  chain-1))

(defun nchain-dot-product (chain-1 chain-2 &key (test #'eql))
  "Computes the dot product of the two chains. 
chain-1 may be modified and returned."
  ;; this implementation relies on delete-if testing each element exactly once:
  ;; See Hyperspec 17.2.1 and 17.2.2:
  ;; "considered iteratively against each element Ei of a sequence S"
  ;; This does not really explicitly document that behaviour, but nothing 
  ;;   else makes sense.
  (error "Chain-destruction is deprecated so far.")
  (delete-if #'(lambda (entry)
		 (let ((other (assoc (car entry) chain-2 :test test)))
		   (when other
		     ;; will be kept so modify
		     (setf (cdr entry) (* (cdr entry) (cdr other))))
		   ;; if no other exists throw away.
		   (not other)))
	     chain-1))

(defun chain* (number chain #|&key test is not needed|# ) 
  "Returns a new chain where chain is multiplied by number.
When number is 1 the specified chain may be returned."
  (cond ((null number) ())
	((zerop number) 
	 (mapcar #'(lambda (entry)
		     `(,(car entry) . ,number))
		 chain))
	((eql number 1)
	 chain)
	(T
	 (mapcar #'(lambda (entry)
		  `(,(car entry) . ,(* number (cdr entry))))
	      chain))))

(defun nchain* (number chain)
  "Multiplies a number into the chain (modifies the chain)."
  (error "Chain-destruction is deprecated so far.")
  (if (eql number 1)
      chain
      (dolist (entry chain chain)
	(setf (cdr entry) (* number (cdr entry))))))
    

;;; Matrix-access:
;;; currently a matrix is an a-list with entries ((row col) . value), hence a chain.

;;; This might be better:
;;; ((a . b) . c) 
;;; two conses:
;;;   [*|*]--->c 
;;;    |  
;;;    v
;;;  [*|*]  
;;;  /   \
;;; a     b
;;;
;;; ((a b) . c)
;;; three conses:
;;;   [*|*]--->c 
;;;    |  
;;;    v
;;;  [*|*]-->[*| ]  
;;;   |       |
;;;   v       v
;;;   a       b



(defun make-matrix-entry (row col value) `((,row ,col) . ,value))
(defun matrix-entry-row  (entry) (caar entry))
(defun (setf matrix-entry-row)  (new-value entry) 
  (error "Matrix-destruction is deprecated so far.")
  (setf (caar entry) new-value))
(defun matrix-entry-col  (entry) (cadar entry))
(defun (setf matrix-entry-col)  (new-value entry) 
  (error "Matrix-destruction is deprecated so far.")
  (setf (cadar entry) new-value))
(defun matrix-entry-value  (entry) (cdr entry))
(defun (setf matrix-entry-value)  (new-value entry) 
  (error "Matrix-destruction is deprecated so far.")
  (setf (cdr entry) new-value))

(defun matrix-entry-test (row-test &optional (col-test row-test))
  #'(lambda(row col entry)
      (and (funcall row-test (matrix-entry-row entry) row)
	   (funcall col-test (matrix-entry-col entry) col))))

(defun matrix-key-test (row-test &optional (col-test row-test))
  "This one is needed to use chain algebra with matrices."
  #'(lambda(row-col-1 row-col-2)
      (and (funcall row-test (first  row-col-1) (first  row-col-2))
	   (funcall col-test (second row-col-1) (second row-col-2)))))


;;; There MUST be a structure to hold these three values for an matrix entry:
;;; Alternatives: 
;;; * an array #(row col value)
;;; * a struct #S(:row row :col col :value value)

;;; There SHOULD be an index to speed up matrix access (hash-tables, AVL-trees, R-trees etc.)
;;; This will be done later:

;;; matrix accessors:
;;; get and set one entry:
(defun matrix-get (matrix row col &key (test #'eql) (col-test test))
   (cdr (find-if (curry (matrix-entry-test test col-test) row col) matrix )))


(defun matrix-nset (matrix row col value &key (test #'eql) (col-test test))
  "Destructively modifies matrix to contain the given value at row and col 
and returns the modified matrix."
  (error "Matrix-destruction is deprecated so far.")
  (let ((pair-test (matrix-entry-test test col-test)))
    (if (null value)
	(delete-if  pair-test  matrix)
	(let ((cell (find-if pair-test matrix )))
	  (if cell
	      (progn (setf (matrix-entry-value cell) value) matrix)
	      (cons (make-matrix-entry row col value) matrix))))))


(defun matrix-set (matrix row col value &key (test #'eql) (col-test test))
  "Returns a new matrix which contains the given value at row and col.

The new matrix may share structure with the specified matrix."
  (let ((pair-test (curry (matrix-entry-test test col-test) row col)))
    (if (null value)
	(remove-if  pair-test  matrix)
	(cons (make-matrix-entry row col value) 
	      (remove-if  pair-test  matrix)))))


;;; When matrix-place is a valid place, then (matrix-get matrix-place row col)  is 
;;; a valid place, too. The matrix object itself will not be modified. 
(define-setf-expander matrix-get (matrix row col &key (test #'eql) &environment env)
  (multiple-value-bind 
	(tmp-vars vals store-vars setter getter)
      (get-setf-expansion matrix env)

    (let ((matrix-var (car store-vars))
	  (row-var    (gensym "ROW"))
	  (col-var    (gensym "COL"))
	  (test-var   (gensym "TST"))
	  (store-var  (gensym "VAL")))
      
      (values
       `(,@tmp-vars ,row-var ,col-var ,test-var)
       `(,@vals     ,row     ,col     ,test)
       `(,store-var)
       `(let ((,matrix-var ,getter))
	  (setq ,matrix-var 
		(matrix-set ,matrix-var ,row-var ,col-var ,store-var :test ,test-var))
	  ,setter
	  ,store-var)
       `(matrix-get ,getter ,row-var ,col-var :test ,test-var)))))


;;; Accessor for a matrix row:
;;; The row index will not be removed from the result and 
;;;   a place for the row-index has to be provided when storing the row:
;;;
;;; (matrix-get-row '(((1 1) . 11) ((1 2) . 12) ((2 2) . 22)) 1)
;;; => (((1 1) . 11) ((1 2) . 12))
;;; OR=> (((1 2) . 12) ((1 1) . 11))
;;;
;;; (matrix-set-row '(((1 1) . 11) ((1 2) . 12) ((2 2) . 22)) 
;;;                  2
;;;                 '(((x 1) . 21) ((3 2) . 32)))
;;; => (((2 1) . 21) ((2 2) . 32) ((1 1) . 11) ((1 2) . 12))
;;; OR=> any other permutation of the above list.
;;;
;;; When P is a place where the matrix can be stored, then the following
;;; equivalence holds:
;;;
;;;  (setf P (matrix-set-row P r-idx row))
;;;  == (setf (matrix-get-row P r-idx) row)

(defun matrix-get-row (matrix row-idx &key (test #'eql))
  "Returns a matrix wich consists of all row entries which have 
the specified row index row-idx."
  (remove row-idx matrix
	  :test (complement test)
	  :key #'matrix-entry-row))

(defun matrix-set-row (matrix row-idx row &key (test #'eql))
  "Returns a matrix where the row entries have the 
specified row index row-idx and returns that new matrix.

The specified row is a list of matrix entries where only the column keys are 
used and the row index will be ignored. 
The consequences are undefined if two different values are specified for one 
column key."
    (nconc   (mapcar #'(lambda (item) 
			 (make-matrix-entry row-idx 
					    (matrix-entry-col item)
					    (matrix-entry-value item)))
		     row )
	     (remove row-idx  matrix
		     :test test
		     :key #'matrix-entry-row)))

(defun matrix-nset-row (matrix row-idx row &key (test #'eql))
  "Destructivley modifies matrix to a matrix where the row entries have the 
specified row index row-idx and returns that new matrix.

The specified row is a list of matrix entries where only the column keys are 
used and the row index will be ignored. Row will not be modified.

The consequences are undefined if two different values are specified for one 
column key."
  (error "Matrix-destruction is deprecated so far.")

  (nconc   (mapcar #'(lambda (item) 
		       (make-matrix-entry row-idx 
					  (matrix-entry-col item)
					  (matrix-entry-value item)))
		   row )
	   (delete row-idx  matrix
		   :test test
		   :key #'matrix-entry-row)))

(defun make-matrix-row (chain &optional row-name)
  (mapcar #'(lambda (item)
	      (make-matrix-entry row-name (car item) (cdr item)))
	  chain))

(defun make-row-chain (row)
  (mapcar #'(lambda (item)
	      (cons (matrix-entry-col item) (matrix-entry-value item)))
	  row))

(define-setf-expander matrix-get-row (matrix row &key (test #'eql) &environment env)
  "This setf expander doesn't modify the matrix istself but the place that holds the matrix."
  (multiple-value-bind 
	(tmp-vars vals store-vars setter getter)
      (get-setf-expansion matrix env)

    (let ((matrix-var (car store-vars))
	  (row-var    (gensym "ROW"))
	  (test-var   (gensym "TST"))
	  (store-var  (gensym "VAL")))
      
      (values
       `(,@tmp-vars ,row-var ,test-var)
       `(,@vals     ,row     ,test)
       `(,store-var)
       `(let ((,matrix-var ,getter))
	  (setq ,matrix-var 
		(matrix-set-row ,matrix-var ,row-var ,store-var :test ,test-var))
	  ,setter
	  ,store-var)
       `(matrix-get-row ,getter ,row-var :test ,test-var)))))


;;; Acessor for a matrix column:
(defun matrix-get-col (matrix col-idx &key (test #'eql))
  "Returns a matrix wich consists of all column entries which have 
the specified column index col-idx."
  (remove col-idx matrix
	  :test (complement test)
	  :key #'matrix-entry-col))

(defun matrix-set-col (matrix col-idx col &key (test #'eql))
  "Returns a matrix where the column entries have the 
specified column index col-idxw.

The specified column stays unchanged."
    (nconc   (mapcar #'(lambda (item) 
			 (make-matrix-entry (matrix-entry-row item)
					    col-idx
					    (matrix-entry-value item)))
		     col )
	     (remove col-idx matrix
		     :test test
		     :key #'matrix-entry-col)))

(defun matrix-nset-col (matrix col-idx col &key (test #'eql))
  "Destructivley modifies matrix to a matrix where the column entries have the 
specified column index col-idx and returns that new matrix.

The specified column stays unchanged."
  (error "Matrix-destruction is deprecated so far.")

  (nconc   (mapcar #'(lambda (item) 
		       (make-matrix-entry (matrix-entry-row item)
					  col-idx
					  (matrix-entry-value item)))
		   col )
	   (delete col-idx matrix
		   :test test
		   :key #'matrix-entry-col)))

(defun make-matrix-col (chain &optional name)
  "Convwerts a chain into a matrix-column with an arbitrary column name."
  (mapcar #'(lambda (item)
	      (make-matrix-entry (car item) 
				 name
				 (cdr item)))
	  chain))

(defun make-col-chain (col)
  "Convwerts a matrix column into a chain by extracting the row-indexes."
  (mapcar #'(lambda (item)
	      (cons (matrix-entry-row item) 
		    (matrix-entry-value item)))
	  col))

(define-setf-expander matrix-get-col (matrix col &key (test #'eql) &environment env)
  (multiple-value-bind 
	(tmp-vars vals store-vars setter getter)
      (get-setf-expansion matrix env)

    (let ((matrix-var (car store-vars))
	  (col-var    (gensym "COL"))
	  (test-var   (gensym "TST"))
	  (store-var  (gensym "VAL")))
      
      (values
       `(,@tmp-vars ,col-var ,test-var)
       `(,@vals     ,col     ,test)
       `(,store-var)
       `(let ((,matrix-var ,getter))
	  (setq ,matrix-var 
		(matrix-set-col ,matrix-var ,col-var ,store-var :test ,test-var))
	  ,setter
	  ,store-var)
       `(matrix-get-col ,getter ,col-var :test ,test-var)))))

(defun matrix-copy (matrix)
  (mapcar #'(lambda (entry)
	      (make-matrix-entry (matrix-entry-row entry)
				 (matrix-entry-col entry)
				 (matrix-entry-value entry)))
	  matrix))


(defun matrix-function (matrix &key (test #'eql))
  (rcurry (curry #'matrix-get matrix) :test test))



(defun direct-product (chain-1 chain-2 &key (test #'eql))
  (mapcan #'(lambda (entry-1)
	      (let ((entry-2 (assoc entry-1 chain-2 :test test)))
		(when (cdr entry-2)
		  `((,(car entry-1) . ,(* (cdr entry-1) (cdr entry-2)))))))
	  chain-1))


(defun map-matrix (function matrix)

  "Maps function over all entries in matrix. and returns 
A matrix with all matrix entries returned by the function.

The function must always return a list of matrix entries to be 
collected into the resulting matrix. No algebraic addition is performed.

The consequences are undefied if function returns anything else 
a valid matrix entry list. Or if it returns \"duplicate\" entries.

\(Note that every element in the empty list is a valid matrix entry)."

  (mapcan function matrix))

(defun map-matr (function matrix)

  "Maps function over all entries in matrix, and returns matrix."

  (mapc function matrix)
  matrix)
