;;;;  dgesv-test.lisp
;;;;  In fact not a test any more but a loader/wrapper for dgesv and other 
;;;;  lapapck routines.
;;;;
;;;;  Copyright (C) 2010  Norbert Paul
;;;;
;;;;  This program is free software: you can redistribute it and/or modify
;;;;  it under the terms of the GNU General Public License as published by
;;;;  the Free Software Foundation, either version 3 of the License, or
;;;;  (at your option) any later version.
;;;;
;;;;  This program is distributed in the hope that it will be useful,
;;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;;  GNU General Public License for more details.
;;;;
;;;;  You should have received a copy of the GNU General Public License
;;;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;;  Contact: norbert.paul@kit.edu
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;  Purpose
;;;;  =======
;;;;
;;;;  DGESV computes the solution to a real system of linear equations
;;;;     A * X = B,
;;;;  where A is an N-by-N matrix and X and B are N-by-NRHS matrices.
;;;;
;;;;  The LU decomposition with partial pivoting and row interchanges is
;;;;  used to factor A as
;;;;     A = P * L * U,
;;;;  where P is a permutation matrix, L is unit lower triangular, and U is
;;;;  upper triangular.  The factored form of A is then used to solve the
;;;;  system of equations A * X = B.
;;;;
;;;;  Arguments
;;;;  =========
;;;;
;;;;  N       (input) INTEGER
;;;;          The number of linear equations, i.e., the order of the
;;;;          matrix A.  N >= 0.
;;;;
;;;;  NRHS    (input) INTEGER
;;;;          The number of right hand sides, i.e., the number of columns
;;;;          of the matrix B.  NRHS >= 0.
;;;;
;;;;  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
;;;;          On entry, the N-by-N coefficient matrix A.
;;;;          On exit, the factors L and U from the factorization
;;;;          A = P*L*U; the unit diagonal elements of L are not stored.
;;;;
;;;;  LDA     (input) INTEGER
;;;;          The leading dimension of the array A.  LDA >= max(1,N).
;;;;
;;;;  IPIV    (output) INTEGER array, dimension (N)
;;;;          The pivot indices that define the permutation matrix P;
;;;;          row i of the matrix was interchanged with row IPIV(i).
;;;;
;;;;  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
;;;;          On entry, the N-by-NRHS matrix of right hand side matrix B.
;;;;          On exit, if INFO = 0, the N-by-NRHS solution matrix X.
;;;;
;;;;  LDB     (input) INTEGER
;;;;          The leading dimension of the array B.  LDB >= max(1,N).
;;;;
;;;;  INFO    (output) INTEGER
;;;;          = 0:  successful exit
;;;;          < 0:  if INFO = -i, the i-th argument had an illegal value
;;;;          > 0:  if INFO = i, U(i,i) is exactly zero.  The factorization
;;;;                has been completed, but the factor U is exactly
;;;;                singular, so the solution could not be computed.
;;;;
;;;;  =====================================================================


(eval-when (:compile-toplevel :load-toplevel)
  (require '#:ASDF)
  (asdf:operate 'asdf:load-source-op 'lapack-dgesv))

(defpackage :laroutines 
            (:use :cl)
            (:export :lapacksolve 
		     :solve 
		     :ddot
		     :depsilon
		     :multiply))

(in-package :laroutines)

(defun lapacksolve (mat y)
  "Solves the equation system Mat*x = y using lapack's dgesv
If the matrix is singular an error is signalled.
The specified matrix mat and vector y are passed unchanged to dgesv and, 
hence,  must be one-dimensional arrays. Mat must be in column-mayor form.
"
(let* ((n (length y))
       (ipiv (make-array n :element-type 'fortran-to-lisp::integer4)) )
  (multiple-value-bind 
	(arg1 arg2 arg3 arg4 arg5 arg6 arg7 info)
      (lapack:dgesv n 1 mat n ipiv y n 0)
    (declare (ignore arg1 arg2 arg3 arg4 arg5 arg6 arg7))
    (cond ((zerop info) y)
	  ((minusp info) 
	   (error "LAPACK:DGESV dislikes the ~D-th argument." (- info)))
	  (T
	   (error "The matrix is singular."))))))


(defun matrix2lapack (mat)
  (let* ((n (first (array-dimensions mat)))
	 (m (second (array-dimensions mat)))
	 (offset 0)
	 (result (make-array (* n m) :element-type 'double-float)))

    (dotimes (col n) 
      (dotimes (row n)
	(setf (aref result (+ offset row)) 
	      (coerce (aref mat row col) 'double-float)))
      (incf offset n))

    result ))

	   

(defun solve (mat y)
  "Solves the equation system Mat*x = y using lapack's dgesv
If the matrix is singular an error is signalled.
The specified matrix mat and vector y are saved and will not be changed.
The matrix must be a two-dimensional array in standard row-mayor form.
The matrix and y entries must be coercible to double-float.
"
  (declare (type  (array * 2) mat)
	   (type  (array * 1) y)
	   (ftype (array double-float)))

  (let* ((n      (length y))
         (lamat  (make-array (* n n) :element-type 'double-float))
	 (ipiv   (make-array n :element-type 'fortran-to-lisp::integer4))
         (result (make-array n :element-type 'double-float))
	 (offset 0))

    (declare (type fixnum n offset)
	     (type (array double-float 1) lamat result)
	     (type (array fortran-to-lisp::integer4 1) ipiv) )

    (dotimes (col n) 
      (setf (aref result col)
	    (coerce (aref y col) 'double-float))
      (dotimes (row n)
	(setf (aref lamat (+ offset row)) 
	      (coerce (aref mat row col) 'double-float)))
      (incf offset n))
    (multiple-value-bind 
	  (arg1 arg2 arg3 arg4 arg5 arg6 arg7 info)

	;; The call to DGESV:
	(lapack:dgesv n 1 lamat n ipiv result n 0)

      (declare (ignore arg1 arg2 arg3 arg4 arg5 arg6 arg7)
	       (type fixnum info))

      (cond ((zerop info) result)
	    ((minusp info) 
	     (error "LAPACK:DGESV dislikes the ~D-th argument." (- info)))
	    (T
	     (error "The matrix is singular."))))))



(defun ddot (x y)

  (declare (type  (simple-array * 1) x y))
  (declare (ftype (function ((simple-array double-float 1) 
			     (simple-array double-float 1)) 
			    double-float) 
		  ddot))

  (let* ((n (min (length x) (length y)))
	 (result (blas:ddot n x 1 y 1)))

    (declare (type fortran-to-lisp::integer4 n)
	     (type double-float result))
    
    result )
)


(defun depsilon ()
  "Returns the machine epsilon for double floats computed by LAPACK."
  (nth-value 0 (lapack:dlamch "E"))
)

(defun multiply (mat vec &optional (result
				    (make-array (first (array-dimensions mat))
						:element-type 'double-float)))
  "Multiplies matrix with vector. 
If a result vector is specified the result is stored into that vector."

  (let ((n (min (length result)
		(first  (array-dimensions mat))))
	(m (min (second (array-dimensions mat))
		(length vec)))
	(acc 0d0))
	
    (dotimes (row n)
      (dotimes (col m)
	(incf acc (* (aref mat row col)
		     (aref vec col))))
      (setf (aref result row) acc   
	    acc 0d0))
    
    result )
)